-- racecartool
-- wiki query script

-- Copyright (c) 2007 Stefan Scheffler

-- Permission is hereby granted, free of charge, to any person
-- obtaining a copy of this software and associated documentation
-- files (the "Software"), to deal in the Software without
-- restriction, including without limitation the rights to use,
-- copy, modify, merge, publish, distribute, sublicense, and/or sell
-- copies of the Software, and to permit persons to whom the
-- Software is furnished to do so, subject to the following
-- conditions:

-- The above copyright notice and this permission notice shall be
-- included in all copies or substantial portions of the Software.

-- THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
-- EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
-- OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
-- NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
-- HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
-- WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
-- FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
-- OTHER DEALINGS IN THE SOFTWARE.

-- known issues:
-- prints loads of 404 errors for dead links
-- depth > 2 untested


local HTTP = dofile("httplib.lua")

-- max search depth
local MAXDEPTH = 2
-- number of concurrent requests
local MAXREQUESTS = 5
local vlist = {}
local vlist_depth = 0

local valid_sections = {
	DEFINITION = true,
	DESCRIPTION = true,
	ARGUMENTS = true,
	RETURNS = true,
	EXAMPLE = true,
	VALUE = true,
	DETAILS = true
}

local defaultformat = { 
	"DEFINITION",
	"DESCRIPTION", 
	"ARGUMENTS", 
	"RETURNS",
	"VALUE",
	"EXAMPLE"
}


-- some helpers
local function upperify(tab)
	result = {}
	for i,v in ipairs(tab) do
		result[i] = string.upper(v)
	end
	return result
end

local function must_download(depth)
	return (depth > vlist_depth) and true 
end

local function update_vlist(new_vlist, depth)
	vlist = new_vlist
	vlist_depth = depth
end

local function fix_depth(depth)
	local new_vlist_depth = (vlist_depth > 0) and vlist_depth or 1 
	local num_depth = tonumber(depth)

	if not num_depth or num_depth < 1 then
		return (vlist_depth > 0) and vlist_depth or 1
	elseif num_depth > MAXDEPTH then
		return MAXDEPTH
	else
		return num_depth
	end
end

-- unused
local function merge_lists(list1, depth1, list2, depth2)
	local new_list = {}
	
	for i,v in pairs(list1) do
		new_list[i] = v
	end
	
	for i,v in pairs(list2) do
		new_list[i] = v
	end
	
	local new_depth = (depth2 > depth1) and depth2 or depth1
	return new_list, new_depth
end


-- filter out the interesting bits out of downloaded pages
-- store result in a table
local function parse(page)
	local category
	local sectionname

	local list = {}
	local entry = {}
	local section = {}

	for line in page:gmatch("[^\r\n]+") do
		if line:gsub(" ", ""):gmatch("==([^=]*)==$")() then
			category = line:gsub(" ", ""):gmatch("==([^=]*)==$")()
		elseif line:gsub(" ", ""):gmatch("===([^=]*)===$")() then
			local entryname = line:gsub(" ", ""):gmatch("===([^=]*)===$")()

			entry = {}
			entry.sections = {}
			entry.category = category

			list[entryname] = entry
		-- too much nesting
		elseif valid_sections[(line:gmatch("'''([^']-):?'''")() or ""):upper()] then
			section = {}
			sectionname = line:gmatch("'''([^']-):?'''")():upper()
			entry.sections[sectionname] = section
		else
			if category == "Tables" and
			   sectionname == "DETAILS" and
			   line:gmatch("%[%[(.*)%]%]")() then
				entry.tablepage = line:gmatch("%[%[(.*)%]%]")()
			end
			
			local filtered_line = line:gsub("<.->", ""):gsub("'''", "\"")

			if #filtered_line > 0 then
				table.insert(section, filtered_line)
			end
		end	
	end

	return list
end

-- request wiki pages
-- walk recursively through the wiki till depth and requests are 0
-- depth is decremented each recursion level starting from given depth
-- requests represents the number of requests that haven't returned yet
-- prefix is prepended to each entry in subpages (should use tree instead)
-- when finished call cb

-- possible optimisation: incremental updates
-- keep representation of previously requested tree of pages around
-- if requested depth > tree depth => start requesting from leafes
-- storing result in a tree would simplify this
local function request(depth, cb, data)
	local done = false
	local requests = 1
	local list = {}
	local requested_depth = depth


	local queue = {}
	
	local function queue_push(request)
		table.insert(queue, 1, request)
	end
	
	local function queue_pop()
		local request = queue[#queue]
		table.remove(queue)
		return request
	end

	local function request2(title, prefix, depth)
		local http = HTTP.new()
		
		http.timeout = 10

		local function http_cb(response)
			requests = requests - 1

			if response.status == 200 then
				for i,v in pairs(parse(response.body.get())) do
					list[prefix..i] = v

					if depth > 1 and v.category == "Tables" and v.tablepage then
						queue_push({v.tablepage, i..".", depth - 1})
					end
				end
			else
				if response.url then
					console_print("download of "..response.url.path.." failed! "..tostring(response.status))
				else
					console_print("download failed! "..tostring(response.status))
				end
			end

			if #queue > 0 then
				for i=1,(MAXREQUESTS - requests) do
					if #queue == 0 then
						break 
					end
					requests = requests + 1

					local request = queue_pop()
					request2(unpack(request))
				end
			elseif requests <= 0 then
				if cb then cb(list, requested_depth - depth + 1, data) end
			end
		end

		http.GET.add("title", title)
		http.GET.add("action", "raw")
		http.urlopen("http://www.vo-wiki.com/racecar/index.php", http_cb)
	end

	print("downloading..")
	request2("API_Index", "", depth)
end


-- print entry name from list
local function print_entry(list, name)
	local section

	if not list[name] then
		print("no entry called "..tostring(name))
		return
	end

	print("|--------\n|"..name)
	for _,v in ipairs(format) do
		section = list[name].sections[v]
		if section and #section > 0 then
			print("|"..v..":")
			for _,w in ipairs(section) do
				print("|"..w)
			end
		end
	end
end

-- find entry in list
-- full determines whether to do full text search
local function find(list, str, full)
	local matches = {}

	if not (str or list) then 
		return
	end

	for i in pairs(list) do
		if i:find(str) then 
			table.insert(matches, i) 
		elseif full then
			for _,v in pairs(list[i].sections) do
				local found = false
				for _,w in ipairs(v) do
					if w:find(str) then found = true end
				end
				if found then
					table.insert(matches, i)
					break
				end
			end
		end
	end
	return matches
end


-- command handlers
local function find_cmd(full, args)
	-- print entry if 1 match
	-- otherwise print list of matches
	local function find_cb(list, depth)
		local matches = find(list, args[1], full)

		if #matches == 1 then
			print_entry(list, matches[1])
		elseif #matches > 1 then
			print(tostring(#matches).." matches")
			
			table.sort(matches)
			for _,v in ipairs(matches) do
				print(v)
			end
		else
			print("no matches")
		end	
		update_vlist(list, depth)
	end

	if args then
		local depth = fix_depth(args[2])
		if must_download(depth) then
			request(depth, find_cb)
		else
			find_cb(vlist, vlist_depth)
		end
	else
		print("missing argument")
	end
end

RegisterUserCommand("rctfind", find_cmd, false)
RegisterUserCommand("rctfindfull", find_cmd, true)

local function print_cmd(_, args)
	local function cb(list, depth)
		print_entry(list, args[1])
		update_vlist(list, depth)
	end

	if args then
		local depth = fix_depth(args[2])
		if must_download(depth) then
			request(depth, cb)
		else
			print_entry(vlist, args[1])
		end
	else
		print("missing argument")
	end	
end

RegisterUserCommand("rctprint", print_cmd)


local function check_format(formatting)
	if not formatting or #formatting == 0 then
		return false
	end

	for _,v in ipairs(upperify(formatting)) do
		if not valid_sections[v] then
			print("invalid section name: "..tostring(v))
			return false
		end
	end

	return true
end

local function format_cmd(_, formatting)
	local up_formatting = upperify(formatting or {})

	if check_format(up_formatting) then
		gkini.WriteString("racecartool", "format", spickle(up_formatting))
		format = up_formatting
	else
		print("rctformat requires a list of section names")

		local sections_string = ""
		for i in pairs(valid_sections) do
			sections_string = sections_string.." "..tostring(i)
		end

		print("valid sections are:"..sections_string)
		
		sections_string = ""
		for _,v in ipairs(format) do
			sections_string = sections_string.." "..tostring(v)
		end
		print("current format is:"..sections_string)
		
		print("example: /rctformat definition description arguments returns example")
	end
end

RegisterUserCommand("rctformat", format_cmd)

-- settings
local function load_format()
	local formatting = upperify(unspickle(gkini.ReadString("racecartool", "format", "")) or {})

	format = check_format(formatting) and formatting or defaultformat 
end
load_format()